@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/python
 
 ## Copyright (C) 2007, 2008, 2009, 2010 Red Hat, Inc.
 ## Author: Tim Waugh <twaugh@redhat.com>
@@ -30,7 +30,7 @@ import pynotify
 import time
 import locale
 import gettext
-import installdriver
+import cupshelpers.installdriver
 from gettext import gettext as _
 DOMAIN="system-config-printer"
 gettext.textdomain (DOMAIN)
@@ -364,12 +364,14 @@ if __name__ == '__main__':
                 pass
 
         try:
-            installdriver.PrinterDriversInstaller(bus)
+            cupshelpers.installdriver.set_debugprint_fn (debugprint)
+            cupshelpers.installdriver.PrinterDriversInstaller(bus)
         except Exception, e:
             try:
                 print >> sys.stderr, ("%s: failed to start "
                                       "PrinterDriversInstaller service: %s" %
                                       (PROGRAM_NAME, e))
+                pass
             except:
                 pass
 
