@@ -173,11 +173,13 @@ class _IPPConnectionThread(threading.Thread):
 
     def _auth (self, prompt, conn=None, method=None, resource=None):
         def prompt_auth (prompt):
+            gtk.gdk.threads_enter ()
             if conn == None:
                 self._auth_handler (prompt, self._conn)
             else:
                 self._auth_handler (prompt, self._conn, method, resource)
 
+            gtk.gdk.threads_leave ()
             return False
 
         if self._auth_handler == None:
@@ -190,7 +192,9 @@ class _IPPConnectionThread(threading.Thread):
     def _reply (self, result):
         def send_reply (handler, result):
             if not self._destroyed:
+                gtk.gdk.threads_enter ()
                 handler (self._conn, result)
+                gtk.gdk.threads_leave ()
             return False
 
         if not self._destroyed and self._reply_handler:
@@ -199,7 +203,9 @@ class _IPPConnectionThread(threading.Thread):
     def _error (self, exc):
         def send_error (handler, exc):
             if not self._destroyed:
+                gtk.gdk.threads_enter ()
                 handler (self._conn, exc)
+                gtk.gdk.threads_leave ()
             return False
 
         if not self._destroyed and self._error_handler:
@@ -437,6 +443,8 @@ class _IPPAuthOperation:
                                      parent=conn.parent)
 
         d.set_prompt (prompt)
+        if self._user == None:
+            self._user = cups.getUser()
         d.set_auth_info ([self._user, ''])
         d.field_grab_focus ('password')
         d.set_keep_above (True)
