@@ -105,7 +105,9 @@ class _PK1AsyncMethodCall:
             return
 
         if str (error) == '':
+            gtk.gdk.threads_enter ()
             self._client_reply_handler (self._conn, self._unpack_fn (*args))
+            gtk.gdk.threads_leave ()
             self._destroy ()
             return
 
@@ -118,7 +120,9 @@ class _PK1AsyncMethodCall:
 
         if exc.get_dbus_name () == CUPS_PK_NEED_AUTH:
             exc = cups.IPPError (cups.IPP_NOT_AUTHORIZED, 'pkcancel')
+            gtk.gdk.threads_enter ()
             self._client_error_handler (self._conn, exc)
+            gtk.gdk.threads_leave ()
             self._destroy ()
             return
 
@@ -153,7 +157,10 @@ class _WriteToTmpFile:
     def __init__ (self, kwds, reply_handler, error_handler):
         self._reply_handler = reply_handler
         self._error_handler = error_handler
-        (tmpfd, tmpfname) = tempfile.mkstemp ()
+
+        # Create the temporary file in /tmp to ensure that
+        # cups-pk-helper-mechanism is able to write to it.
+        (tmpfd, tmpfname) = tempfile.mkstemp (dir="/tmp")
         os.close (tmpfd)
         self._filename = tmpfname
         debugprint ("Created tempfile %s" % tmpfname)
