@@ -1,8 +1,8 @@
 ## system-config-printer
 
-## Copyright (C) 2006, 2007, 2008, 2009 Red Hat, Inc.
+## Copyright (C) 2006, 2007, 2008, 2009, 2010 Red Hat, Inc.
 ## Copyright (C) 2006 Florian Festi <ffesti@redhat.com>
-## Copyright (C) 2006, 2007, 2008, 2009 Tim Waugh <twaugh@redhat.com>
+## Copyright (C) 2006, 2007, 2008, 2009, 2010 Tim Waugh <twaugh@redhat.com>
 
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -45,7 +45,7 @@ class Printer:
             os.unlink(self._ppd)
 
     def __repr__ (self):
-        return "<cupshelpers.Printer \"%s\">" % self.name
+        return "<cupshelpers.Printer \"%s\" at 0x%x>" % (self.name, id (self))
 
     def _expand_flags(self):
 
@@ -327,7 +327,12 @@ class Printer:
         """
         ret = []
         try:
-            jobs = self.connection.getJobs ()
+            try:
+                r = ['job-id', 'job-printer-uri', 'job-name']
+                jobs = self.connection.getJobs (requested_attributes=r)
+            except TypeError:
+                # requested_attributes requires pycups 1.9.50
+                jobs = self.connection.getJobs ()
         except cups.IPPError:
             return ret
 
@@ -357,7 +362,13 @@ class Printer:
         """
         ret = []
         try:
-            jobs = self.connection.getJobs (which_jobs='completed')
+            try:
+                r = ['job-id', 'job-printer-uri', 'job-state']
+                jobs = self.connection.getJobs (which_jobs='completed',
+                                                requested_attributes=r)
+            except TypeError:
+                # requested_attributes requires pycups 1.9.50
+                jobs = self.connection.getJobs (which_jobs='completed')
         except cups.IPPError:
             return ret
 
@@ -464,7 +475,7 @@ def parseDeviceID (id):
         if piece.find(":") == -1:
             continue
         name, value = piece.split(":",1)
-        id_dict[name] = value
+        id_dict[name.strip ()] = value.strip()
     if id_dict.has_key ("MANUFACTURER"):
         id_dict.setdefault("MFG", id_dict["MANUFACTURER"])
     if id_dict.has_key ("MODEL"):
@@ -506,6 +517,10 @@ class Device:
 
         self.id_dict = parseDeviceID (self.id)
 
+        s = uri.find("serial=")
+        if s != -1 and not self.id_dict.get ('SN',''):
+            self.id_dict['SN'] = uri[s + 7:]
+
     def __repr__ (self):
         return "<cupshelpers.Device \"%s\">" % self.uri
 
@@ -535,6 +550,10 @@ class Device:
                 return 1
             if self.type == "hp" or self.type == "hpfax":
                 return -1
+            if other.type == "dnssd":
+                return 1
+            if self.type == "dnssd":
+                return -1
             if other.type == "usb":
                 return 1
             if self.type == "usb":
