@@ -22,7 +22,9 @@
 
 import cups
 from .cupshelpers import parseDeviceID
+import itertools
 import string
+import time
 import locale
 import os.path
 import re
@@ -31,6 +33,8 @@ from . import _debugprint, set_debugprint_fn
 __all__ = ['ppdMakeModelSplit',
            'PPDs']
 
+_RE_version_numbers = re.compile (" v\d(?:\d*\.\d+)?(?: |$)")
+
 def ppdMakeModelSplit (ppd_make_and_model):
     """
     Split a ppd-make-and-model string into a canonical make and model pair.
@@ -171,17 +175,25 @@ def ppdMakeModelSplit (ppd_make_and_model):
         make = "HP"
         makel = "hp"
 
-    # HP PPDs give NickNames like:
+    # HP PostScript PPDs give NickNames like:
     # *NickName: "HP LaserJet 4 Plus v2013.111 Postscript (recommended)"
-    # Find the version number.
+    # Find the version number and truncate at that point.  But beware,
+    # other model names can legitimately look like version numbers,
+    # e.g. Epson PX V500.
+    # Truncate only if the version number has only one digit, or a dot
+    # with digits before and after.
     modell = model.lower ()
     v = modell.find (" v")
-    if v != -1 and (model[v + 2].isdigit () or
-                    (model[v + 2] == '.' and
-                     model[v + 3].isdigit ())):
-        # Truncate at that point.
-        model = model[:v]
-        modell = modell[:v]
+    if v != -1:
+        # Look for " v" followed by a digit, optionally followed by more
+        # digits, a dot, and more digits; and terminated by a space of the
+        # end of the line.
+        vmatch = _RE_version_numbers.search (modell)
+        if vmatch:
+            # Found it -- truncate at that point.
+            vstart = vmatch.start ()
+            modell = modell[:vstart]
+            model = model[:vstart]
 
     for suffix in [" hpijs",
                    " foomatic/",
@@ -199,6 +211,9 @@ def ppdMakeModelSplit (ppd_make_and_model):
                    " zxs",              # hpcups
                    " pcl3",             # hpcups
                    "_bt",
+                   " pcl",              # Canon CQue
+                   " ufr ii",           # Canon UFR II
+                   " br-script3",       # Brother PPDs
                    ","]:
         s = modell.find (suffix)
         if s != -1:
@@ -224,24 +239,30 @@ def ppdMakeModelSplit (ppd_make_and_model):
     model = model.strip ()
     return (make, model)
 
+def _singleton (x):
+    """If we don't know whether getPPDs() or getPPDs2() was used, this
+    function can unwrap an item from a list in either case."""
+    if isinstance (x, list):
+        return x[0]
+    return x
+
 # Some drivers are just generally better than others.
 # Here is the preference list:
 DRIVER_TYPE_DOWNLOADED_NOW = 5
 DRIVER_TYPE_FOOMATIC_RECOMMENDED_NON_POSTSCRIPT = 8
 DRIVER_TYPE_VENDOR = 10
 DRIVER_TYPE_FOOMATIC_RECOMMENDED_POSTSCRIPT = 15
+DRIVER_TYPE_HPCUPS = 16
 DRIVER_TYPE_FOOMATIC_HPIJS_ON_HP = 17
-DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED = 20
-DRIVER_TYPE_GUTENPRINT_NATIVE = 25
 DRIVER_TYPE_SPLIX = 27
 DRIVER_TYPE_FOOMATIC_PS = 30
 DRIVER_TYPE_FOOMATIC_HPIJS = 40
-DRIVER_TYPE_HPCUPS = 45
+DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED = 42
+DRIVER_TYPE_GUTENPRINT_NATIVE = 45
 DRIVER_TYPE_FOOMATIC_GUTENPRINT_SIMPLIFIED = 50
 DRIVER_TYPE_FOOMATIC_GUTENPRINT = 60
 DRIVER_TYPE_FOOMATIC = 70
 DRIVER_TYPE_CUPS = 80
-DRIVER_TYPE_FOOMATIC_GENERIC = 90
 DRIVER_TYPE_3RD_PARTY_NONFREE = 95
 DRIVER_DOES_NOT_WORK = 999
 def _getDriverType (ppdname, ppds=None):
@@ -252,7 +273,7 @@ def _getDriverType (ppdname, ppds=None):
         return DRIVER_TYPE_SPLIX
     if ppdname.find ("hpcups") != -1:
         info = ppds.getInfoFromPPDName (ppdname)
-        make_model = info.get ('ppd-make-and-model', '')
+        make_model = _singleton (info.get ('ppd-make-and-model', ''))
         if make_model.find ("plugin") != -1:
             return DRIVER_TYPE_3RD_PARTY_NONFREE
         else:
@@ -260,13 +281,11 @@ def _getDriverType (ppdname, ppds=None):
     if (ppdname.find (":") == -1 and
         ppdname.find ("/cups-included/") != -1):
         return DRIVER_TYPE_CUPS
-    if ppdname.startswith ("foomatic:"):
+    if ppdname.startswith ("foomatic") and ppdname.find (":") != -1:
         # Foomatic (generated) -- but which driver?
-        if ppdname.find ("Generic")!= -1:
-            return DRIVER_TYPE_FOOMATIC_GENERIC
         if ppds != None:
             info = ppds.getInfoFromPPDName (ppdname)
-            device_id = info.get ('ppd-device-id', '')
+            device_id = _singleton (info.get ('ppd-device-id', ''))
             id_dict = parseDeviceID (device_id)
             drv = id_dict.get ('DRV', '')
             drvfields = dict()
@@ -276,7 +295,8 @@ def _getDriverType (ppdname, ppds=None):
                 key = field[0]
                 drvfields[key] = field[1:]
 
-            ppd_make_and_model = info.get ('ppd-make-and-model', '')
+            ppd_make_and_model = _singleton (info.get ('ppd-make-and-model',
+                                                       ''))
             if (drvfields.get ('R', '0') == '1' or
                 ppd_make_and_model.find ("(recommended)") != -1):
                 if ppd_make_and_model.find ("Postscript") != -1:
@@ -293,12 +313,17 @@ def _getDriverType (ppdname, ppds=None):
                 return DRIVER_TYPE_FOOMATIC_GUTENPRINT_SIMPLIFIED
             return DRIVER_TYPE_FOOMATIC_GUTENPRINT
         return DRIVER_TYPE_FOOMATIC
-    if ppdname.find ("gutenprint") != -1:
+    if ppdname.startswith ("gutenprint"):
         if (ppdname.find ("/simple") != -1 or
             ppdname.find (".sim-") != -1):
             return DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED
         else:
             return DRIVER_TYPE_GUTENPRINT_NATIVE
+    if ppdname.startswith ("ijsgutenprint"):
+        if ppdname.find ("-simplified") != -1:
+            return DRIVER_TYPE_FOOMATIC_GUTENPRINT_SIMPLIFIED
+        else:
+            return DRIVER_TYPE_FOOMATIC_GUTENPRINT
     if ppdname.find ("-hpijs") != -1:
         if ppdname.find ("hpijs-rss") == -1:
             return DRIVER_TYPE_FOOMATIC_HPIJS
@@ -325,6 +350,7 @@ class PPDs:
         """
         @type ppds: dict
         @param ppds: dict of PPDs as returned by cups.Connection.getPPDs()
+        or cups.Connection.getPPDs2()
 
         @type language: string
 	@param language: language name, as given by the first element
@@ -348,7 +374,7 @@ class PPDs:
         to_remove = []
         for ppdname, ppddict in self.ppds.iteritems ():
             try:
-                natural_language = ppddict['ppd-natural-language']
+                natural_language = _singleton (ppddict['ppd-natural-language'])
             except KeyError:
                 continue
 
@@ -372,7 +398,7 @@ class PPDs:
         # which unfortunately then appears as manufacturer Raw and
         # model Queue.  Use 'Generic' for this model.
         if self.ppds.has_key ('raw'):
-            makemodel = self.ppds['raw']['ppd-make-and-model']
+            makemodel = _singleton (self.ppds['raw']['ppd-make-and-model'])
             if not makemodel.startswith ("Generic "):
                 self.ppds['raw']['ppd-make-and-model'] = "Generic " + makemodel
 
@@ -441,7 +467,7 @@ class PPDs:
             return ppdnamelist
 
         dict = self.getInfoFromPPDName (ppdnamelist[0])
-        make_model = dict['ppd-make-and-model']
+        make_model = _singleton (dict['ppd-make-and-model'])
         mfg, mdl = ppdMakeModelSplit (make_model)
         def getDriverTypeWithBias (x, mfg):
             t = _getDriverType (x, ppds=self)
@@ -552,6 +578,8 @@ class PPDs:
 	@returns: an integer,string pair of (status,ppd-name)
 	"""
         _debugprint ("\n%s %s" % (mfg, mdl))
+        orig_mfg = mfg
+        orig_mdl = mdl
         self._init_ids ()
 
         # Start with an empty result list and build it up using
@@ -648,6 +676,72 @@ class PPDs:
                 status = self.STATUS_GENERIC_DRIVER
                 ppdnamelist = generic
 
+        # What about the CMD field of the Device ID?  Some devices
+        # have optional units for page description languages, such as
+        # PostScript, and they will report different CMD strings
+        # accordingly.
+        #
+        # By convention, if a PPD contains a Device ID with a CMD
+        # field, that PPD can only be used whenever any of the
+        # comma-separated words in the CMD field appear in the
+        # device's ID.
+        # (See Red Hat bug #630058).
+        #
+        # We'll do that check now, and any PPDs that fail
+        # (e.g. PostScript PPD for non-PostScript printer) can be
+        # eliminated from the list.
+        #
+        # The reason we don't do this check any earlier is that we
+        # don't want to eliminate PPDs only to have the fuzzy matcher
+        # add them back in afterwards.
+        if id_matched and len (commandsets) > 0:
+            failed = set()
+            for ppdname in ppdnamelist:
+                ppd = self.ppds[ppdname]
+                ppd_device_id = _singleton (ppd.get ('ppd-device-id'))
+                if not ppd_device_id:
+                    continue
+
+                ppd_device_id_dict = parseDeviceID (ppd_device_id)
+                ppd_cmd_field = ppd_device_id_dict["CMD"]
+                if not ppd_cmd_field:
+                    continue
+
+                usable = False
+                for pdl in ppd_cmd_field:
+                    if pdl in commandsets:
+                        usable = True
+                        break
+
+                if not usable:
+                    failed.add (ppdname)
+
+            _debugprint ("Removed %s due to CMD mis-match" % failed)
+            ppdnamelist = list (set (ppdnamelist) - failed)
+
+        # If the Device ID matched but the DES field is different,
+        # eliminate those PPDs too.
+        if id_matched and description:
+            _debugprint ("Checking DES field")
+            inexact = set()
+            for ppdname in ppdnamelist:
+                if (ppdname.find ("hpijs") != -1 or
+                    ppdname.find ("hpcups") != -1):
+                    continue
+                ppddict = self.ppds[ppdname]
+                id = _singleton (ppddict['ppd-device-id'])
+                if not id:
+                    continue
+                # Fetch description field.
+                id_dict = parseDeviceID (id)
+                if id_dict["DES"] != description:
+                    inexact.add (ppdname)
+
+            exact = set (ppdnamelist).difference (inexact)
+            _debugprint ("discarding: %s" % inexact)
+            if len (exact) >= 1:
+                ppdnamelist = list (exact)
+
         if not ppdnamelist:
             status = self.STATUS_NO_DRIVER
             fallbacks = ["textonly.ppd", "postscript.ppd"]
@@ -671,27 +765,6 @@ class PPDs:
                 _debugprint ("No fallback available; choosing any")
                 ppdnamelist = [self.ppds.keys ()[0]]
 
-        if id_matched:
-            _debugprint ("Checking DES field")
-            inexact = set()
-            if description:
-                for ppdname in ppdnamelist:
-                    if ppdname.find ("hpijs"):
-                        continue
-                    ppddict = self.ppds[ppdname]
-                    id = ppddict['ppd-device-id']
-                    if not id:
-                        continue
-                    # Fetch description field.
-                    id_dict = parseDeviceID (id)
-                    if id_dict["DES"] != description:
-                        inexact.add (ppdname)
-
-            exact = set (ppdnamelist).difference (inexact)
-            _debugprint ("discarding: %s" % inexact)
-            if len (exact) >= 1:
-                ppdnamelist = list (exact)
-
         # We've got a set of PPDs, any of which will drive the device.
         # Now we have to choose the "best" one.  This is quite tricky
         # to decide, so let's sort them in order of preference and
@@ -707,7 +780,7 @@ class PPDs:
                 cmd = reduce (lambda x, y: x + ","+ y, commandsets)
             except TypeError:
                 cmd = ""
-            id = "MFG:%s;MDL:%s;" % (mfg, mdl)
+            id = "MFG:%s;MDL:%s;" % (orig_mfg, orig_mdl)
             if cmd:
                 id += "CMD:%s;" % cmd
             if description:
@@ -862,15 +935,15 @@ class PPDs:
             return get ("PostScript Printer")
         elif (("pclxl" in cmdsets) or ("pcl-xl" in cmdsets) or
               ("pcl6" in cmdsets) or ("pcl 6 emulation" in cmdsets)):
-            return get ("PCL 6/PCL XL Printer")
+            return get ("PCL 6/PCL XL Printer", "PCL Laser Printer")
         elif "pcl5e" in cmdsets:
-            return get ("PCL 5e Printer")
+            return get ("PCL 5e Printer", "PCL Laser Printer")
         elif "pcl5c" in cmdsets:
-            return get ("PCL 5c Printer")
+            return get ("PCL 5c Printer", "PCL Laser Printer")
         elif ("pcl5" in cmdsets) or ("pcl 5 emulation" in cmdsets):
-            return get ("PCL 5 Printer")
+            return get ("PCL 5 Printer", "PCL Laser Printer")
         elif "pcl" in cmdsets:
-            return get ("PCL 3 Printer")
+            return get ("PCL 3 Printer", "PCL Laser Printer")
         elif (("escpl2" in cmdsets) or ("esc/p2" in cmdsets) or
               ("escp2e" in cmdsets)):
             return get ("ESC/P Dot Matrix Printer")
@@ -880,11 +953,12 @@ class PPDs:
         if self.makes:
             return
 
+        tstart = time.time ()
         makes = {}
         lmakes = {}
         lmodels = {}
         for ppdname, ppddict in self.ppds.iteritems ():
-            ppd_make_and_model = ppddict['ppd-make-and-model']
+            ppd_make_and_model = _singleton (ppddict['ppd-make-and-model'])
             (make, model) = ppdMakeModelSplit (ppd_make_and_model)
             lmake = make.lower ()
             lmodel = model.lower ()
@@ -906,6 +980,7 @@ class PPDs:
         self.makes = makes
         self.lmakes = lmakes
         self.lmodels = lmodels
+        _debugprint ("init_makes: %.3fs" % (time.time () - tstart))
 
     def _init_ids (self):
         if self.ids:
@@ -913,9 +988,7 @@ class PPDs:
 
         ids = {}
         for ppdname, ppddict in self.ppds.iteritems ():
-            if not ppddict.has_key ('ppd-device-id'):
-                continue
-            id = ppddict['ppd-device-id']
+            id = _singleton (ppddict.get ('ppd-device-id'))
             if not id:
                 continue
 
@@ -925,10 +998,8 @@ class PPDs:
 
             bad = False
             if len (lmfg) == 0:
-                _debugprint ("Missing MFG field for %s" % ppdname)
                 bad = True
             if len (lmdl) == 0:
-                _debugprint ("Missing MDL field for %s" % ppdname)
                 bad = True
             if bad:
                 continue
@@ -990,7 +1061,14 @@ def _self_test(argv):
     except IOError:
         f = open (picklefile, "w")
         c = cups.Connection ()
-        cupsppds = c.getPPDs ()
+        try:
+            cupsppds = c.getPPDs2 ()
+            print "Using getPPDs2()"
+        except AttributeError:
+            # Need pycups >= 1.9.52 for getPPDs2
+            cupsppds = c.getPPDs ()
+            print "Using getPPDs()"
+
         pickle.dump (cupsppds, f)
 
     ppds = PPDs (cupsppds)
@@ -1043,7 +1121,7 @@ def _self_test(argv):
          "DES:Canon iP3000;VER:1.09;STA:10;FSI:03;", 1, "Canon PIXMA iP3000"),
         ("MFG:HP;MDL:Deskjet 5400 series;CMD:MLC,PCL,PML,DW-PCL,DESKJET,DYN;"
          "1284.4DL:4d,4e,1;CLS:PRINTER;DES:5440;",
-         1, "HP DeskJet (5440)|(5550)"), # foomatic-db-hpijs used to say 5440
+         1, "HP DeskJet (5440|5550)"), # foomatic-db-hpijs used to say 5440
         ("MFG:Hewlett-Packard;MDL:HP LaserJet 3390;"
          "CMD:PJL,MLC,PCL,POSTSCRIPT,PCLXL;",
          0, "HP LaserJet 3390"),
@@ -1056,7 +1134,6 @@ def _self_test(argv):
         ("MFG:New;MDL:Unknown PCL5c Printer;CMD:PCL5c;", 2, "Generic PCL 5c"),
         ("MFG:New;MDL:Unknown PCL5 Printer;CMD:PCL5;", 2, "Generic PCL 5"),
         ("MFG:New;MDL:Unknown PCL3 Printer;CMD:PCL;", 2, "Generic PCL"),
-        ("MFG:New;MDL:Unknown ESC/P Printer;CMD:ESCP2E;", 2, "Generic ESC/P"),
         ("MFG:New;MDL:Unknown Printer;", 100, None),
         ]
 
@@ -1075,7 +1152,9 @@ def _self_test(argv):
             success = True
         else:
             if status == max_status_code:
-                match = re.match (modelre, ppddict['ppd-make-and-model'], re.I)
+                match = re.match (modelre,
+                                  _singleton (ppddict['ppd-make-and-model']),
+                                  re.I)
                 success = match != None
             else:
                 success = False
@@ -1086,7 +1165,7 @@ def _self_test(argv):
             result = "*** FAIL ***"
 
         print "%s: %s %s (%s)" % (result, id_dict["MFG"], id_dict["MDL"],
-                                  ppddict['ppd-make-and-model'])
+                                  _singleton (ppddict['ppd-make-and-model']))
         all_passed = all_passed and success
 
     if not all_passed:
