@@ -338,6 +338,7 @@ class Monitor:
 
             self.update_timer = gobject.timeout_add (200,
                                                      self.get_notifications)
+            debugprint ("Deferred get_notifications by 200ms")
             return False
 
         debugprint ("get_notifications")
@@ -363,6 +364,9 @@ class Monitor:
                     return False
 
                 self.watcher.cups_ipp_error (self, e, m)
+                if e == cups.IPP_FORBIDDEN:
+                    return False
+
                 return True
         except RuntimeError:
             cups.setUser (user)
@@ -418,6 +422,12 @@ class Monitor:
                 continue
 
             # Job events
+            if not nse.startswith ("job-"):
+                # Some versions of CUPS give empty
+                # notify-subscribed-event attributes (STR #3608).
+                debugprint ("Unhandled nse %s" % repr (nse))
+                continue
+
             jobid = event['notify-job-id']
             if (nse == 'job-created' or
                 (nse == 'job-state-changed' and
@@ -486,6 +496,7 @@ class Monitor:
             interval = notifications['notify-get-interval']
             t = gobject.timeout_add_seconds (interval,
                                              self.get_notifications)
+            debugprint ("Next notifications fetch in %ds" % interval)
             self.update_timer = t
 
         return False
@@ -535,14 +546,19 @@ class Monitor:
 
         try:
             self.sub_id = c.createSubscription ("/", events=events)
+            debugprint ("Created subscription %d, events=%s" % (self.sub_id,
+                                                                repr (events)))
         except cups.IPPError, (e, m):
             self.watcher.cups_ipp_error (self, e, m)
 
         cups.setUser (user)
 
-        self.update_timer = gobject.timeout_add_seconds (MIN_REFRESH_INTERVAL,
-                                                         self.get_notifications)
-        debugprint ("Created subscription %d" % self.sub_id)
+        if self.sub_id != -1:
+            self.update_timer = gobject.timeout_add_seconds (
+                MIN_REFRESH_INTERVAL,
+                self.get_notifications)
+            debugprint ("Next notifications fetch in %ds" %
+                        MIN_REFRESH_INTERVAL)
 
         if self.monitor_jobs:
             jobs = self.jobs.copy ()
@@ -737,8 +753,10 @@ class Monitor:
             gobject.source_remove (self.update_timer)
 
         self.update_timer = gobject.timeout_add (200, self.get_notifications)
+        debugprint ("Next notifications fetch in 200ms (update called)")
 
     def handle_dbus_signal(self, *args):
+        debugprint ("D-Bus signal from CUPS... calling update")
         self.update ()
         if not self.received_any_dbus_signals:
             self.received_any_dbus_signals = True
