@@ -21,6 +21,7 @@
 
 import dbus
 import gtk
+import cupshelpers
 
 import asyncconn
 from debug import debugprint
@@ -31,19 +32,39 @@ def set_gettext_function (fn):
     _ = fn
 
 class PPDsLoader:
-    def __init__ (self, callback, device_id=None, parent=None,
-                  host=None, encryption=None):
+    """
+    1. If PackageKit support is available, and this is a local server,
+    try to use PackageKit to install relevant drivers.  We do this
+    because we can only make the right choice about the "best" driver
+    when the full complement of drivers is there to choose from.
+
+    2. Fetch the list of available drivers from CUPS.
+
+    3. If Jockey is available, and there is no appropriate driver
+    available, try to use Jockey to install one.
+
+    4. If Jockey was able to install one, fetch the list of available
+    drivers again.
+    """
+    def __init__ (self, callback, device_id=None, device_uri=None, parent=None,
+                  host=None, encryption=None, language=None):
         debugprint ("+%s" % self)
         self._callback = callback
         self._device_id = device_id
+        self._device_uri = device_uri
         self._parent = parent
         self._host = host
         self._encryption = encryption
+        self._language = language
 
         self._installed_files = []
         self._conn = None
         self._ppds = None
-
+        self._ppdsmatch_result = (cupshelpers.ppds.PPDs.STATUS_NO_DRIVER, '')
+        self._jockey_queried = False
+        self._local_cups = (self._host == None or
+                            self._host == "localhost" or
+                            self._host[0] == '/')
         try:
             self._bus = dbus.SessionBus ()
         except:
@@ -62,16 +83,13 @@ class PPDsLoader:
 
         self._dialog.connect ("response", self._dialog_response)
 
-        if (device_id and self._bus and
-
-            # Only try to install packages if we are configuring the
-            # local CUPS server.
-            (self._host == None or
-             self._host == "localhost" or
-             self._host[0] == '/')):
+        self._dialog.show_all ()
+        if self._local_cups and self._device_id and self._bus:
+            self._devid_dict = cupshelpers.parseDeviceID (self._device_id)
+            self._gpk_device_id = "MFG:%s;MDL:%s;" % (self._devid_dict["MFG"],
+                                                      self._devid_dict["MDL"])
             self._query_packagekit ()
         else:
-            self._dialog.show_all ()
             self._query_cups ()
 
     def __del__ (self):
@@ -92,70 +110,22 @@ class PPDsLoader:
     def get_ppds (self):
         return self._ppds
 
+    def get_ppdsmatch_result (self):
+        return self._ppdsmatch_result
+
     def _dialog_response (self, dialog, response):
         self._call_callback (None)
 
-    def _query_packagekit (self):
-        debugprint ("Asking PackageKit to install drivers")
-        try:
-            xid = self._parent.window.xid
-        except:
-            xid = 0
-
-        try:
-            obj = self._bus.get_object ("org.freedesktop.PackageKit",
-                                        "/org/freedesktop/PackageKit")
-            proxy = dbus.Interface (obj, "org.freedesktop.PackageKit.Modify")
-            proxy.InstallPrinterDrivers (xid, [self._device_id],
-                                         "hide-finished",
-                                         reply_handler=self._packagekit_reply,
-                                         error_handler=self._packagekit_error,
-                                         timeout=3600)
-        except Exception, e:
-            debugprint ("Failed to talk to PackageKit: %s" % e)
-            if self._dialog:
-                self._dialog.show_all ()
-                self._query_jockey ()
-
-    def _packagekit_reply (self):
-        debugprint ("Got PackageKit reply")
-        if self._dialog:
-            self._dialog.show_all ()
-            self._query_jockey ()
-
-    def _packagekit_error (self, exc):
-        debugprint ("Got PackageKit error: %s" % exc)
-        if self._dialog:
-            self._dialog.show_all ()
-            self._query_jockey ()
-
-    def _query_jockey (self):
-        debugprint ("Asking Jockey to install drivers")
-        try:
-            obj = self._bus.get_object ("com.ubuntu.DeviceDriver", "/GUI")
-            jockey = dbus.Interface (obj, "com.ubuntu.DeviceDriver")
-            r = jockey.search_driver ("printer_deviceid:%s" % devid,
-                                      reply_handler=self._jockey_reply,
-                                      error_handler=self._jockey_error,
-                                      timeout=3600)
-        except Exception, e:
-            self._jockey_error (e)
-
-    def _jockey_reply (self, result):
-        debugprint ("Got Jockey result: %s" % repr (result))
-        self._installed_files = result[1]
-        self._query_cups ()
-
-    def _jockey_error (self, exc):
-        debugprint ("Got Jockey error: %s" % exc)
-        self._query_cups ()
-
     def _query_cups (self):
         debugprint ("Asking CUPS for PPDs")
-        c = asyncconn.Connection (host=self._host, encryption=self._encryption,
-                                  reply_handler=self._cups_connect_reply,
-                                  error_handler=self._cups_error)
-        self._conn = c
+        if (not self._conn):
+            c = asyncconn.Connection (host=self._host,
+                                      encryption=self._encryption,
+                                      reply_handler=self._cups_connect_reply,
+                                      error_handler=self._cups_error)
+            self._conn = c
+        else:
+            self._cups_connect_reply(self._conn, None)
 
     def _cups_connect_reply (self, conn, UNUSED):
         if conn != self._conn:
@@ -171,8 +141,32 @@ class PPDsLoader:
             conn.destroy ()
             return
 
+        ppds = cupshelpers.ppds.PPDs (result, language=self._language)
+        self._ppds = ppds
+        self._need_requery_cups = False
+        if self._device_id:
+            self._ppdsmatch_result = ppds.\
+                getPPDNameFromDeviceID (self._devid_dict["MFG"],
+                                        self._devid_dict["MDL"],
+                                        self._devid_dict["DES"],
+                                        self._devid_dict["CMD"],
+                                        self._device_uri,
+                                        self._installed_files)
+
+            (status, ppdname) = self._ppdsmatch_result
+            if (self._bus and
+                status != ppds.STATUS_SUCCESS and
+                not self._jockey_queried and
+                self._local_cups):
+                # Try to install packages using jockey if
+                # - there's no appropriate driver (PPD) locally available
+                # - we are configuring local CUPS server
+                self._jockey_queried = True
+                self._query_jockey ()
+                return
+
         conn.destroy ()
-        self._ppds = result
+        self._conn = None
         self._call_callback (None)
 
     def _cups_error (self, conn, exc):
@@ -181,9 +175,78 @@ class PPDsLoader:
             return
 
         conn.destroy ()
+        self._conn = None
         self._ppds = None
         self._call_callback (exc)
 
+    def _query_packagekit (self):
+        debugprint ("Asking PackageKit to install drivers")
+        try:
+            xid = self._parent.window.xid
+        except:
+            xid = 0
+
+        try:
+            obj = self._bus.get_object ("org.freedesktop.PackageKit",
+                                        "/org/freedesktop/PackageKit")
+            proxy = dbus.Interface (obj, "org.freedesktop.PackageKit.Modify")
+            resources = [self._gpk_device_id]
+            interaction = "hide-finished"
+            debugprint ("Calling InstallPrinterDrivers (%s, %s, %s)" %
+                        (repr (xid), repr (resources), repr (interaction)))
+            proxy.InstallPrinterDrivers (dbus.UInt32 (xid),
+                                         resources, interaction,
+                                         reply_handler=self._packagekit_reply,
+                                         error_handler=self._packagekit_error,
+                                         timeout=3600)
+        except Exception, e:
+            debugprint ("Failed to talk to PackageKit: %s" % e)
+            if self._dialog:
+                self._dialog.show_all ()
+                self._query_cups ()
+
+    def _packagekit_reply (self):
+        debugprint ("Got PackageKit reply")
+        self._need_requery_cups = True
+        if self._dialog:
+            self._dialog.show_all ()
+            self._query_cups ()
+
+    def _packagekit_error (self, exc):
+        debugprint ("Got PackageKit error: %s" % exc)
+        if self._dialog:
+            self._dialog.show_all ()
+            self._query_cups ()
+
+    def _query_jockey (self):
+        debugprint ("Asking Jockey to install drivers")
+        try:
+            obj = self._bus.get_object ("com.ubuntu.DeviceDriver", "/GUI")
+            jockey = dbus.Interface (obj, "com.ubuntu.DeviceDriver")
+            r = jockey.search_driver ("printer_deviceid:%s" % self._device_id,
+                                      reply_handler=self._jockey_reply,
+                                      error_handler=self._jockey_error,
+                                      timeout=3600)
+        except Exception, e:
+            self._jockey_error (e)
+
+    def _jockey_reply (self, conn, result):
+        debugprint ("Got Jockey result: %s" % repr (result))
+        try:
+            self._installed_files = result[1]
+        except:
+            self._installed_files = []
+        self._query_cups ()
+
+    def _jockey_error (self, exc):
+        debugprint ("Got Jockey error: %s" % exc)
+        if self._need_requery_cups:
+            self._query_cups ()
+        else:
+            self._conn.destroy ()
+            self._conn = None
+            self._call_callback (None)
+
     def _call_callback (self, exc):
         if self._callback:
             self._callback (self, exc)
